import gzip
import json
import pickle
import re
import sys
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from os import environ
from pathlib import Path
from typing import Mapping, Sequence, Type, TypeVar, cast

from rich.table import Table
from rich.text import Text
from yaralyzer.util.helpers.interaction_helper import ask_to_proceed

from epstein_files.documents.config.doc_cfg import Metadata
from epstein_files.documents.config.manual_config import create_configs
from epstein_files.documents.document import Document, DocType
from epstein_files.documents.documents.doc_types_mixin import DocTypesMixin
from epstein_files.documents.documents.search_result import SearchResult
from epstein_files.documents.doj_file import DojFile
from epstein_files.documents.email import EMAILERS_TO_ALWAYS_TRUNCATE, Email
from epstein_files.documents.emails.constants import UNINTERESTING_EMAILERS
from epstein_files.documents.emails.dropsite_email import DropsiteEmail
from epstein_files.documents.emails.emailers import CONFIGURED_NON_ENTITIES, ENTITIES_DICT
from epstein_files.documents.emails.multi_email_files import split_up_multi_email_files
from epstein_files.documents.json_file import JsonFile
from epstein_files.documents.messenger_log import MSG_REGEX, MessengerLog
from epstein_files.documents.messenger_log_pdf import IMESSAGE_PDF_IDS, MessengerLogPdf
from epstein_files.documents.other_file import OtherFile
from epstein_files.output.rich import TABLE_TITLE_STYLE, console
from epstein_files.people.entity import Entity
from epstein_files.people.person import PEOPLE_BIOS, Person
from epstein_files.util.constant.strings import *
from epstein_files.util.constants import *
from epstein_files.util.env import args, logger
from epstein_files.util.helpers.data_helpers import flatten, json_safe, sort_dict_by_keys, uniquify, uniq_sorted
from epstein_files.util.helpers.file_helper import all_txt_paths, doj_txt_paths, extract_file_id, file_size_str
from epstein_files.util.timer import Timer

# Lists of properties to copy into duplicate documents (will be preceded with 'extracted_')
PROPS_TO_COPY = ['author', 'timestamp']
EMAIL_PROPS_TO_COPY = ['recipients']
SLOW_FILE_SECONDS = 1.0


@dataclass
class EpsteinFiles(DocTypesMixin):
    """
    Attributes:
        file_paths (list[Path]): paths to Epstein related text documents
        documents (list[Document]): all parsed Documents except the emails with was_split_up flag
        _uninteresting_ccs (list[Name]): names of tangential people who were just CCed once or similar
    """
    file_paths: list[Path] = field(init=False)
    # Derived fields
    _empty_file_ids: set[str] = field(default_factory=set)
    _emailers: list[Person] = field(default_factory=list)
    _uninteresting_ccs: list[Name] = field(default_factory=list)

    def __post_init__(self):
        """Iterate through files and build appropriate objects."""
        self.file_paths = sorted(all_txt_paths(), reverse=True)
        self._documents = self._load_file_paths(self.file_paths)
        self._documents += self._split_up_big_emails()
        self._finalize_data_and_write_to_disk()

    @classmethod
    def get_files(cls, timer: Timer | None = None) -> 'EpsteinFiles':
        """Alternate constructor that reads/writes a pickled version of the data ('timer' arg is for logging)."""
        timer = timer or Timer()

        if args.pickle_path.exists() and not args.overwrite_pickle:
            with gzip.open(args.pickle_path, 'rb') as file:
                epstein_files = pickle.load(file)
                timer_msg = f"Loaded {len(epstein_files.documents):,} documents from '{args.pickle_path}'"
                timer.print_at_checkpoint(f"{timer_msg} ({file_size_str(args.pickle_path)})")

            if args.load_new:
                epstein_files.load_new_files()
            elif args.reload_doj:
                epstein_files.reload_doj_files()

            return epstein_files

        logger.warning(f"Building new cache file, this will take a few minutes...")
        epstein_files = EpsteinFiles()
        num_synthetic_cfgs = len([c for c in CONFIGS_BY_ID.values() if c.is_synthetic])
        timer.print_at_checkpoint(f'Processed {len(epstein_files.file_paths):,} files, {num_synthetic_cfgs} synthetic configs')
        return epstein_files

    @property
    def counterparties_dict(self) -> dict[Name, list[Name]]:
        """Keys are names, values are lists of all the people who sent/received communication with that person."""
        return sort_dict_by_keys({p.name: p.counterparties for p in self.emailers})

    @property
    def documents(self) -> Sequence[Document]:
        """Overloads mixing @property to exclude split up big files."""
        return [d for d in self._documents if not (isinstance(d, Email) and d._was_split_up)]

    @property
    def emails_with_attachments(self) -> list[Email]:
        return [e for e in self.emails if e.attached_docs]

    @property
    def emailers(self) -> list[Person]:
        """All the people who sent or received an email."""
        return self._emailers

    @property
    def uninteresting_emailers(self) -> list[Name]:
        """Emailers whom we don't want to print a separate section for because they're just CCed."""
        if '_uninteresting_emailers' not in vars(self):
            self._uninteresting_emailers = uniq_sorted(UNINTERESTING_EMAILERS + self._uninteresting_ccs)

        return self._uninteresting_emailers

    def docs_for(self, name: Name) -> list[Document]:
        """All documents with `name` as the author or a recipient (not just someone who is mentioned)."""
        docs = flatten([fxn(name) for fxn in [self.emails_for, self.imessage_logs_for, self.other_files_for]])
        return Document.sort_by_timestamp(Document.uniquify(docs))

    def email_author_counts(self) -> dict[Name, int]:
        """Returns dict counting up how many emails were written by each person."""
        return {
            person.name: len(person.unique_emails_by)
            for person in self.emailers if len(person.unique_emails_by) > 0
        }

    def email_authors_to_device_signatures(self) -> dict[str, set[str]]:
        """Mapping of authors to all the device signatures identified in their history."""
        signatures = defaultdict(set)

        for email in [e for e in self.unique_emails if e.sent_from_device]:
            signatures[email.author_str].add(email.sent_from_device)

        return signatures

    def email_device_signatures_to_authors(self) -> dict[str, set[str]]:
        """Mapping of device signatures to all the users who ever signed an email with them."""
        signatures = defaultdict(set)

        for email in [e for e in self.unique_emails if e.sent_from_device]:
            signatures[email.sent_from_device].add(email.author_str)

        return signatures

    def email_recipient_counts(self) -> dict[Name, int]:
        """Returns dict with names as keys and number of emails received as values."""
        return {p.name: len(p.unique_emails_to) for p in self.emailers if len(p.unique_emails_to) > 0}

    def email_signature_substitution_counts(self) -> dict[str, int]:
        """Return the number of times an email signature was replaced with "<...snipped...>" for each author."""
        substitution_counts = defaultdict(int)

        for email in self.emails:
            for name, num_replaced in email.signature_substitution_counts.items():
                substitution_counts[name] += num_replaced

        return substitution_counts

    def emails_by(self, author: Name) -> list[Email]:
        """All emails sent by `author` (including dupes!)."""
        return Document.sort_by_timestamp([e for e in self.emails if e.author == author])

    def emails_for(self, name: Name) -> list[Email]:
        """All emails to or from 'name' sorted chronologically (including dupes!)."""
        emails = self.emails_by(name) + self.emails_to(name)
        emails += [e for e in self.emails if name and name == e._config.show_with_name]  # Add show_with_name emails

        if len(emails) == 0:
            logger.warning(f"No emails found for '{name}'")

        return Document.sort_by_timestamp(Document.uniquify(emails))

    def emails_to(self, name: Name) -> list[Email]:
        """All `Email`s sent to `name` (including dupes!)."""
        if name is None:
            emails = [e for e in self.emails if len(e.recipients) == 0 or None in e.recipients]
        else:
            emails = [e for e in self.emails if name in e.recipients]

        return Document.sort_by_timestamp(emails)

    def get_ids(self, ids: list[str], rebuild: bool = False) -> Sequence[Document]:
        """Get `Document` objects for `file_ids`. If `rebuild` is True then rebuild `Document` from .txt file."""
        docs = [self.docs_by_id[id] for id in ids]

        if len(docs) != len(ids):
            logger.error(f"{len(ids)} file IDs provided but only {len(docs)} documents found! ids: {ids}")

        return [doc.reload() if rebuild else doc for doc in docs]

    def get_id(self, file_id: str, rebuild: bool = False, required_type: Type[DocType] = Document) -> DocType:
        """Singular ID version of `get_ids()` but with option to require a type of document subclass."""
        doc = self.get_ids([file_id], rebuild)[0]

        if not isinstance(doc, required_type):
            raise ValueError(f"No {required_type.__name__} found for {file_id} (found {doc})")

        return doc

    def grep_documents(self, pattern: re.Pattern | str, names: list[Name] | None = None) -> list[SearchResult]:
        """Find documents whose text matches `pattern` optionally limited to only docs involving `name`)."""
        documents = [d for d in self._documents if (not names) or d.author in names]
        results: list[SearchResult] = []

        for doc in documents:
            if doc.is_duplicate:
                continue

            lines = doc.lines_matching(pattern)

            if args.min_line_length:
                lines = [line for line in lines if len(line.line) > args.min_line_length]

            if len(lines) > 0:
                results.append(SearchResult(doc, lines))

        return results

    def imessage_logs_for(self, name: Name) -> list[MessengerLog]:
        """Return `MessengerLog` objects where Epstein's counterparty is `name`."""
        return [log for log in self.imessage_logs if name == log.author]

    def json_metadata(self) -> str:
        """Create a JSON string containing metadata for all the files."""
        def _sorted_metadata(docs: Sequence[Document]) -> list[Metadata]:
            return [json_safe(d.metadata) for d in Document.sort_by_id(docs)]

        metadata = {
            'files': {
                Email.__name__: _sorted_metadata(self.emails),
                JsonFile.__name__: _sorted_metadata(self.json_files),
                MessengerLog.__name__: _sorted_metadata(self.imessage_logs),
                OtherFile.__name__: _sorted_metadata(self.non_json_other_files),
            },
            'people': {name: bio.plain for name, bio in PEOPLE_BIOS.items()},
        }

        return json.dumps(metadata, indent=4, sort_keys=True)

    def load_new_files(self) -> None:
        """Load any new files detected in the hierarchy."""
        new_docs = self._load_file_paths(self._new_files())

        if not new_docs:
            logger.warning(f"No new files found, doing nothing.")
            return
        elif (cfgs := create_configs(new_docs)):  # Exit if new configs were created
            console.print(f"\n{len(cfgs)} config objects created, exiting...", style='bold yellow1')
            sys.exit()

        self._finalize_data_and_write_to_disk(new_docs)

    def other_files_for(self, name: Name) -> list[OtherFile]:
        """Get files with author `name` or that are marked `show_with_name`."""
        if name is None:
            return []
        else:
            return [f for f in self.other_files if name in [f.author, f._config.show_with_name]]

    def overview_table(self) -> Table:
        """Table showing file counts by type."""
        title = Text('Files Overview ', TABLE_TITLE_STYLE)
        title.append('(last updated: ', 'dim').append(datetime.now().date().isoformat(), 'cyan').append(')', 'dim')
        table = Document.files_summary_table(title, 'File Type')
        table.add_row('Emails', *Document.file_summary_row(self.emails))
        table.add_row('iMessage Logs', *Document.file_summary_row(self.imessage_logs))
        table.add_row('JSON Data', *Document.file_summary_row(self.json_files, True))
        table.add_row('Other', *Document.file_summary_row(self.non_json_other_files))
        return table

    def person_objs(self, names: list[Name]) -> list[Person]:
        """Construct Person objects for a list of names."""
        return [
            Person(
                name=name,
                _documents=self.docs_for(name),
                is_interesting=False if name in self.uninteresting_emailers else None
            )
            for name in names
        ]

    def reload_doj_files(self) -> None:
        """Reload only the DOJ PDF extracts (keep HOUSE_OVERSIGHT stuff unchanged)."""
        def doj_file_counts_str():
            return f"(have {len(self.all_doj_files)}, {len(self.doj_files)} non-email)"

        timer = Timer()
        logger.warning(f"Reloading all DOJ files {doj_file_counts_str()}...")
        self._finalize_data_and_write_to_disk(self._load_file_paths(doj_txt_paths()))
        timer.print_at_checkpoint(f"Reloaded {len(self.doj_files)} DOJ files {doj_file_counts_str()}")

    def repair_ids(self, ids: list[str]) -> None:
        """Reload the `ids` and save updated pickle file (also loads new files)."""
        ids = uniquify(ids)

        if ids == ['EMAIL']:
            doc_paths = [d.file_path for d in self.emails]
            msg = f"Repairing all {len(doc_paths)} emails..."
        else:
            doc_paths = [d.file_path for d in self._documents if d.file_id in ids]
            msg = f"Repairing {len(ids)} file IDs"

            if len(doc_paths) != len(ids):
                raise RuntimeError(f"{len(ids)} specified but only {len(doc_paths)} Document objects found!")

        if (new_paths := self._new_files()):
            msg +=  f" (also loading {len(new_paths)} new files)"
            doc_paths += new_paths

        logger.warning(msg)
        repaired_docs = self._load_file_paths(doc_paths)

        if environ.get('RESPLIT_BIG_EMAILS'):
            repaired_docs += self._split_up_big_emails()
            logger.warning(f"  (RESPLIT_BIG_EMAILS so now have {len(repaired_docs)} repaired_docs)")

        self._finalize_new_docs_if_approved(repaired_docs)

    def _copy_duplicate_doc_properties(self) -> None:
        """Ensure dupe docs have the properties of the docs they duplicate to capture any repairs, config etc."""
        for doc in self.documents:
            if not doc._config.duplicate_of_id:
                continue

            original = self.get_id(doc._config.duplicate_of_id)
            props_to_copy = (EMAIL_PROPS_TO_COPY + PROPS_TO_COPY) if isinstance(doc, Email) else PROPS_TO_COPY
            props_to_copy = [f"extracted_{prop}" for prop in props_to_copy]

            for field_name in props_to_copy:
                original_prop = getattr(original, field_name)
                duplicate_prop = getattr(doc, field_name)

                if original_prop != duplicate_prop:
                    doc._warn(f"Replacing {field_name} {duplicate_prop} with {original_prop} from duplicated '{original.file_id}'")
                    setattr(doc, field_name, original_prop)

    def _finalize_data_and_write_to_disk(self, new_docs: list[Document] | None = None) -> None:
        """Handle computation of fields related to uninterestingness, relationships between documents, etc."""
        new_docs = new_docs or []

        if new_docs:
            old_num_docs = len(self._documents)
            new_doc_ids = [d.file_info.file_id for d in new_docs]
            self._documents = [d for d in self._documents if d.file_info.file_id not in new_doc_ids]  # Remove existing
            logger.warning(f"Adding {len(new_docs)} Documents (replacing {old_num_docs - len(self._documents)} existing)")
            self._documents += new_docs

        self._set_uninteresting_ccs()
        self._copy_duplicate_doc_properties()
        self._emailers = self.person_objs(flatten([e.participants for e in self.emails]))
        self._find_email_attachments_and_set_is_first_for_user()
        self._documents = Document.sort_by_timestamp(self._documents)
        self._docs_by_id = {doc.file_id: doc for doc in self._documents}
        self._save_to_disk()

    def _find_email_attachments_and_set_is_first_for_user(self) -> None:
        """Add documents with configured `attached_to_email_id` to the `Email`s they're attached to."""
        for email in self.emails:
            email.attached_docs = []  # Remove all attachments before re-finding them in case it's a repair

        email_attachments = [f for f in self.other_files if f._config.attached_to_email_id]
        logger.warning(f"Finding homes for {len(email_attachments)} known email attachments...")

        for attachment in email_attachments:
            email = self.get_id(attachment._config.attached_to_email_id, required_type=Email)
            email.attached_docs.append(attachment)

            if email.is_duplicate:
                raise ValueError(f"Cannot attach {attachment.file_id} to duplicate email {email}")

            # Set the attachment timestamp to that of the email if the attachment has no configured timestamp
            if attachment.timestamp and (attachment.timestamp != email.timestamp) and not attachment._config.timestamp:
                attachment._warn(f"Overwriting '{attachment.timestamp}' with {email}'s timestamp {email.timestamp}")
                attachment.extracted_timestamp = email.timestamp

        # Set the is_persons_first_email flag on the earliest Email we have for each person (sent or received)
        for emailer in self.emailers:
            if emailer.name not in EMAILERS_TO_ALWAYS_TRUNCATE and emailer.unique_emails:
                emailer.unique_emails[0].is_persons_first_email = True

    def _finalize_new_docs_if_approved(self, new_docs: list[Document]) -> None:
        """Same as _finalize_data_and_write_to_disk() but prints new docs and asks for permission."""
        if len(new_docs) < 100:
            console.print(*new_docs)  # Print for user review
            logger.warning(f"Finalizing {len(new_docs)} files...")
        else:
            logger.warning(f"Too many new documents ({len(new_docs)}) to show previews...")

        Document._print_ids(new_docs, 'newly loaded or repaired by EpsteinFiles')
        ask_to_proceed("Looks good?")
        self._finalize_data_and_write_to_disk(new_docs)

    def _load_file_paths(self, file_paths: list[Path]) -> list[Document]:
        """Load a list of file paths into a list of `Document` object subclasses."""
        docs: list[Document] = []

        for file_path in file_paths:
            doc_timer = Timer(decimals=2)
            document = Document(file_path)

            if document.length == 0:
                if document.file_id not in self._empty_file_ids:
                    document._warn(f"Skipping empty file...")
                    self._empty_file_ids.add(document.file_id)

                continue

            cls = document_cls(document)
            # docs.append(cls(file_path, lines=document.lines, text=document.text))
            docs.append(cls(file_path))  # TODO: needs to reload DropsiteEmail
            logger.info(str(docs[-1]))

            if doc_timer.seconds_since_start() > SLOW_FILE_SECONDS:
                doc_timer.print_at_checkpoint(f"Slow file: {docs[-1]} processed")

        return docs

    def _new_files(self) -> list[Path]:
        """Find any files that don't exist in the collection. Has side effect of setting `self.file_paths`."""
        self.file_paths = sorted(all_txt_paths(), reverse=True)
        current_doc_ids = [id for id in self.docs_by_id.keys()] + list(self._empty_file_ids)

        return [
            p for p in self.file_paths
            if extract_file_id(p) not in current_doc_ids
        ]

    def _save_to_disk(self) -> None:
        """Write a pickled version of this `EpsteinFiles` object with all documents etc."""
        with gzip.open(args.pickle_path, 'wb') as file:
            pickle.dump(self, file)
            logger.warning(f"Pickled data to '{args.pickle_path}' ({file_size_str(args.pickle_path)})...")

    def _set_uninteresting_ccs(self) -> None:
        """Extract the recipients of emails configured has having uninteresting CCs or BCCs."""
        for email in self.emails:
            if email._config.has_uninteresting_ccs:
                self._uninteresting_ccs += email.recipients

            if email._config.has_uninteresting_bccs:
                self._uninteresting_ccs += [bcc.lower() for bcc in cast(list[str], email.header.bcc)]

        self._uninteresting_ccs = uniq_sorted(self._uninteresting_ccs)
        logger.info(f"Extracted uninteresting_ccs: {self._uninteresting_ccs}")

    def _split_up_big_emails(self) -> list[Email]:
        """Find the big emails that we want to split up into smaller emails."""
        big_emails = [e for e in self._documents if isinstance(e, Email) and e._was_split_up]

        big_emails = big_emails or [
            *self.emails_by(CHRISTOPHER_DILORIO),
            *[self.get_id(LEON_BLACK_EMAIL_ID, required_type=Email)]
        ]

        new_emails = split_up_multi_email_files(big_emails)
        logger.warning(f"Split up {len(big_emails)} into {len(new_emails)} smaller emails...")
        return new_emails


def document_cls(doc: Document) -> Type[Document]:
    """Find the appropriate `Document` subclass for this file based on the contents."""
    if doc.length == 0:
        return Document
    elif doc.file_info.is_eml_file:
        return DropsiteEmail
    elif doc.text[0] == '{':
        return JsonFile
    elif doc.file_id in IMESSAGE_PDF_IDS:
        return MessengerLogPdf
    elif doc.is_email:
        return Email
    elif doc.file_info.is_doj_file:
        return DojFile
    elif MSG_REGEX.search(doc.text[0:5000]):  # Limit search area to avoid pointless scans of huge files
        return MessengerLog
    else:
        return OtherFile


# Find author/recipients that we configured but that have no Entity so we can suppress warnings about them
# TODO: this populates the var in emailers.py which sucks but there's circular import issues
for cfg in CONFIGS_BY_ID.values():
    for name in cfg.names:
        if name in ENTITIES_DICT or name in CONFIGURED_NON_ENTITIES:
            continue

        CONFIGURED_NON_ENTITIES[name] = Entity(name, is_emailer=False, is_interesting=False)
        log_msg = f"Configured name has no Entity object: {quote(name)}"

        if ',' in name or isinstance(cfg, EmailCfg) and (cfg.has_uninteresting_ccs or cfg.has_uninteresting_bccs):
            cfg._debug_log(log_msg)
        else:
            cfg._log(log_msg)



"""
Small methods for building rich text constructs.
"""
import os
from abc import ABC, abstractmethod
from typing import Sequence, TypeVar

from rich.padding import Padding
from rich.text import Text

from epstein_files.util.constant.strings import QUESTION_MARKS
from epstein_files.util.helpers.string_helper import starred_header

QUESTION_MARKS_TXT = Text(QUESTION_MARKS, style='grey50')
WARNING_STYLE = 'bold black on white'

RAINBOW = [
    'royal_blue1',
    'medium_purple',
    'light_coral',
    'light_slate_gray',
    'dark_goldenrod',
    'wheat4',
    'white',
    'medium_orchid',
    'deep_pink1',
    'navajo_white1',
]

left_indent_padding = lambda num_spaces: (0, 0, 0, num_spaces)
left_indent = lambda obj, num_spaces: Padding(obj, left_indent_padding(num_spaces))
no_bold = lambda style: style.replace('bold', '').strip()
suppress_output_console_kwargs = lambda: {'file': open(os.devnull, "wt")}
vertically_pad = lambda obj, amount = 1: Padding(obj, (amount, 0, amount, 0))


class TextCast(ABC):
    """Abstract class for things that implement a __rich__() method to retun `Text`."""

    @abstractmethod
    def __rich__(self) -> Text:
        pass

Textish = Text | TextCast | str
TextVar = TypeVar('TextVar', bound=Text | str)


def enclose(txt: str | Text, encloser: str = '', encloser_style: str = 'wheat4') -> Text:
    """Surround with something."""
    if not encloser:
        return Text('').append(txt)
    elif len(encloser) != 2:
        raise ValueError(f"'encloser' arg is '{encloser}' which is not 2 characters long")

    enclose_start, enclose_end = (encloser[0], encloser[1])
    return Text('').append(enclose_start, encloser_style).append(txt).append(enclose_end, encloser_style)


def join_non_empty(*txts, sep: str | Text = ' ') -> Text:
    """Join two strings but only if they are not empty."""
    sep = sep if isinstance(sep, Text) else Text(sep)
    return sep.join([t for t in txts if len(t) > 0])


def join_texts(
    txts: Sequence[Textish],
    join: str | Text = ' ',
    encloser: str = '',
    encloser_style: str = 'wheat4',
) -> Text:
    """Join a collection of `Text` objs into one, similar to standard `str.join()`. Skips 0 length stuff."""
    txt = Text('')

    for i, t in enumerate(txts):
        if len((_txt := to_txt(t))) > 0:
            txt.append(join if i >= 1 else '').append(enclose(_txt, encloser, encloser_style))

    return txt


def parenthesize(msg: Textish, parentheses_style: str = '') -> Text:
    return Text('(', style=parentheses_style).append(to_txt(msg)).append(')')


def starred_header_txt(msg: str, num_stars: int = 7, num_spaces: int = 2, style: str = WARNING_STYLE) -> Text:
    """String like '  *** Title Msg ***  '."""
    return Text(starred_header(msg, num_stars, num_spaces), style)


def to_txt(textish: Textish) -> Text:
    if isinstance(textish, Text):
        return textish
    elif isinstance(textish, str):
        return Text(textish)
    else:
        return textish.__rich__()



from epstein_files.documents.config.communication_cfg import skype_log
from epstein_files.documents.config.doc_cfg import DocCfg
from epstein_files.documents.config.email_cfg import EmailCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *

AL_SECKEL_BILL_FIGHT = f"{AL_SECKEL} and Epstein fight about the bill for reputation management services"


REPUTATION_CFGS = [
    DocCfg(id='030426', author=IAN_OSBORNE, note=f"reputation repair proposal citing Michael Milken", date='2011-06-14'),
    DocCfg(id='026582', note=f"Epstein's internet search results at start of reputation repair campaign, maybe from {OSBORNE_LLP}"),
    DocCfg(id='030573', note=f"Epstein's unflattering Google search results, maybe screenshot by {AL_SECKEL} or {OSBORNE_LLP}"),
    DocCfg(id='030875', note=f"Epstein's Wikipedia page", date='2014-02-08'),  # Date is based on tyler shears; seckel was 2010
    DocCfg(id='026583', note=f"Google search results for '{JEFFREY_EPSTEIN}' with analysis ({OSBORNE_LLP}?)"),
    DocCfg(id='029350', note=f"Microsoft Bing search results for Epstein with sex offender at top", attached_to_email_id='EFTA00675542'),
    EmailCfg(id='022203', note=AL_SECKEL_BILL_FIGHT, truncate_to=500),
    EmailCfg(id='022219', note=AL_SECKEL_BILL_FIGHT, truncate_to=2404),

    # DOJ files
    EmailCfg(id='EFTA01830035', note=AL_SECKEL_BILL_FIGHT),
    DocCfg(
        id='EFTA01810372',
        author=TYLER_SHEARS,
        note=f'invoice for reputation management work',
        is_interesting=True,
        attached_to_email_id='EFTA01931256'
    ),
    skype_log('EFTA01217787', recipients=[TYLER_SHEARS, HANNA_TRAFF], is_interesting=True),
    skype_log('EFTA01217703', recipients=[ATHENA_ZELCOVICH, JOSCHA_BACH, LAWRENCE_KRAUSS, TYLER_SHEARS]),
    skype_log('EFTA01217736', recipients=[ATHENA_ZELCOVICH, TYLER_SHEARS]),
]



from dataclasses import dataclass, field
from enum import StrEnum, auto

from rich.text import Text

from epstein_files.documents.documents.categories import Interesting, Neutral, Uninteresting
from epstein_files.documents.config.doc_cfg import DocCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.helpers.data_helpers import without_falsey
from epstein_files.util.helpers.string_helper import join_truthy

class Platform(StrEnum):
    EMAIL = auto()
    IMESSAGE = 'iMessage'
    LETTER = auto()
    SKYPE = 'Skype'
    TEXT_MSG = 'text message'
    WHATSAPP = 'WhatsApp'

PLATFORM_STYLES = {
    Platform.LETTER: 'plum4',
    Platform.SKYPE: 'bright_cyan',
    Platform.WHATSAPP: 'cyan',
}


@dataclass(kw_only=True)
class CommunicationCfg(DocCfg):
    """
    Manual config is always required for MessengerLog author attribution. It's also often needed for Email
    files to handle the terrible OCR text that Congress provided which messes up a lot of the email headers.

    Attributes:
        is_fwded_article (bool, optional): `True` if this is a newspaper article someone fwded. Used to exclude articles from word counting.
        platform (Platform, optional): platform this happened on (e.g. Skype, WhatsApp, etc.)
        recipients (list[Name]): Who received the communication
        recipient_uncertain (bool | str, optional): Optional explanation of why this recipient was attributed, but uncertainly
    """
    is_fwded_article: bool | None = None
    platform: Platform | Literal[''] = ''
    recipients: list[Name] = field(default_factory=list)
    recipient_uncertain: bool | str = ''

    def __post_init__(self):
        super().__post_init__()

        if not isinstance(self.recipients, list):
            raise ValueError(f"{self.id} recipients is not a list: {self.recipients}")

        if self.is_fwded_article:
            self.is_valid_for_name_scan = False

        self.recipients = sort_names(self.recipients)

    @property
    def category_txt(self) -> Text:
        """Overloads superclass to return `self.platform`."""
        if self.platform and self.platform in PLATFORM_STYLES:
            return Text(self.platform.lower(), PLATFORM_STYLES[self.platform])
        else:
            return super().category_txt

    @property
    def complete_description(self) -> str:
        """Add platform to superclass return value."""
        if self.platform == Platform.EMAIL:
            return self.note
        elif self.platform:
            author = f"{self.author} {QUESTION_MARKS}" if self.author and self.author_uncertain else self.author
            recipients = self.recipients_str

            if self.platform in [Platform.SKYPE, Platform.WHATSAPP]:
                description = f"{self.platform} log"
                recipients = join_truthy(author, recipients, ', ')
                recipients_sep = ' of conversation with '
            else:
                description = join_truthy(self.platform, author, ' from ')
                recipients_sep = ' to '

            description = join_truthy(description, recipients, recipients_sep)
            return join_truthy(description, self.note)
        else:
            return super().complete_description

    @property
    def is_of_interest(self) -> bool | None:
        """Fwded articles are not interesting."""
        if self.is_fwded_article and not self.is_interesting:
            return False
        else:
            return super().is_of_interest

    @property
    def names(self) -> list[str]:
        """Names configured for this document. Overloaded in subclass to add recipients."""
        return super().names + without_falsey(self.recipients)

    @property
    def recipients_str(self) -> str:
        return ', '.join([r or UNKNOWN for r in self.recipients])

    def __repr__(self) -> str:
        return super().__repr__()


# TODO: rename MessengerLogCfg
@dataclass(kw_only=True)
class TextCfg(CommunicationCfg):
    """Config for the `MessengerLog` class."""
    platform: Platform = Platform.IMESSAGE

    @property
    def complete_description(self) -> str:
        """Just return the `note` field for iMessage log files."""
        return self.note

    # This is necessary because for some dumb reason @dataclass(repr=False) doesn't cut it
    def __repr__(self) -> str:
        return super().__repr__()


def imessage_log(id: str, **kwargs) -> CommunicationCfg:
    return CommunicationCfg(id=id, platform=Platform.IMESSAGE, **kwargs)


def imessage_screenshot(id: str, **kwargs) -> CommunicationCfg:
    return imessage_log(id=id, note='screenshot(s)', **kwargs)


def skype_log(id: str, **kwargs) -> CommunicationCfg:
    return _communication_app_log(id, Platform.SKYPE, **kwargs)


def whatsapp_log(id: str, **kwargs) -> CommunicationCfg:
    return _communication_app_log(id, Platform.WHATSAPP, **kwargs)


def _communication_app_log(id: str, platform: Platform, **kwargs) -> CommunicationCfg:
    return CommunicationCfg(id=id, platform=platform, show_full_panel=True, **kwargs)



from epstein_files.documents.config.config_builder import fedex_invoice, important_messages_pad, press_release
from epstein_files.documents.config.doc_cfg import DocCfg
from epstein_files.documents.config.communication_cfg import imessage_screenshot, skype_log
from epstein_files.documents.documents.categories import Interesting, Neutral
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *

LUTNICKS_CANTOR = f"Howard Lutnick's {CANTOR_FITZGERALD}"


MISC_CFGS = [
    press_release('015462', 'Nautilus Education', note=f'magazine issue for Epstein foundation {QUESTION_MARKS}', is_interesting=True),
    press_release('029326', EPSTEIN_VI_FOUNDATION, '2013-02-15'),
    press_release('026565', EPSTEIN_VI_FOUNDATION, '2013-02-15', comment=f'maybe a draft of 029326'),
    DocCfg(id='022494', author='DOJ', note=f'Foreign Corrupt Practices Act (FCPA) Resource Guide'),
    DocCfg(id='023096', author=EPSTEIN_VI_FOUNDATION, note=f'blog post', date='2012-11-15'),
    DocCfg(id='027071', author=FEMALE_HEALTH_COMPANY, note=f"brochure requesting donations for female condoms in Uganda"),
    DocCfg(id='027074', author=FEMALE_HEALTH_COMPANY, note=f"pitch deck (USAID was a customer)"),
    DocCfg(id='032735', author=GORDON_GETTY, note=f"on Trump", date='2018-03-20'),  # Dated based on concurrent emails from Getty
    DocCfg(id='025540', author=JEFFREY_EPSTEIN, note=f"rough draft of his side of the story"),
    DocCfg(id='026634', author='Michael Carrier', note=f'comments about an Apollo linked fund "DE Fund VIII"'),
    DocCfg(id='031425', author=SCOTT_J_LINK, note=f'completely redacted email', is_interesting=False),
    DocCfg(id='020447', author='Working Group on Chinese Influence Activities in the US', note=f'Promoting Constructive Vigilance'),
    DocCfg(id='012718', note=f"{CVRA} congressional record", date='2011-06-17'),
    DocCfg(id='019448', note=f"Haitian business investment proposal called Jacmel", attached_to_email_id='019446'),
    DocCfg(id='023644', note=f"interview with Mohammed bin Salman", date='2016-04-25', is_interesting=False),
    DocCfg(
        id='030142',
        author=f"{KATHRYN_RUEMMLER}, {KEN_STARR}",
        date='2016-09-01',
        note=f"mostly empty {JASTA} (Justice Against Sponsors of Terrorism Act) doc referencing suit against Saudis",
    ),
    DocCfg(id='029328', note=f"Rafanelli Events promotional deck", is_interesting=False),
    DocCfg(id='029475', note=f'{VIRGIN_ISLANDS} Twin City Mobile Integrated Health Services (TCMIH) proposal/request for donation', is_interesting=False),
    DocCfg(id='029448', note=f"weird short essay titled 'President Obama and Self-Deception'"),
    press_release(
        '033338',
        NICHOLAS_RIBIS,
        '2000-06-07',
        f"end of {DONALD_TRUMP} & {NICHOLAS_RIBIS} working relationship at Trump's casino",
        is_interesting=True,
    ),

    # DOJ files
    DocCfg(
        id='EFTA00003149',
        date='2016-01-01',
        date_uncertain='complete guess',
        note=f"{LITTLE_SAINT_JAMES} staff list",
        show_full_panel=True,
    ),
    DocCfg(
        id='EFTA00034357',
        author=BUREAU_OF_PRISONS,
        date='2019-08-10',
        note=f"internal message about discovery of Epstein's body",
        background_color='red'
    ),
    DocCfg(
        id='EFTA01063811',
        date='2014-08-07',
        date_uncertain=True,
        note="1201-1300 of 4,285 Apple Address Book entries from one of Epstein's accounts",
        is_interesting=True,
    ),
    DocCfg(id='EFTA01103465', author=BEN_GOERTZEL, date='2013-12-02', note='funding proposal for AI labs in Africa etc.'),
    DocCfg(id='EFTA00006085', author='Broward Center for the Performing Arts', display_text='Front of House Managers Report including ticket sales'),
    DocCfg(id='EFTA00007585', note='various fedex receipts and realtor notes', date='2005-09-20'),
    DocCfg(id='EFTA01193705', note="list of Epstein's known email addresses, internet accounts, cars, boats, airplanes, and telephone numbers"),
    DocCfg(id='EFTA00165515', note="contractor describes Epstein's gun safes", show_full_panel=True),
    DocCfg(id='EFTA00266322', note=f"documents about pitches for non-profits in Australia, including to Effective Altruism"),
    DocCfg(
        id='EFTA00029538',
        note=f"{GHISLAINE_MAXWELL} investigation",
        highlight_quote="the defendant provided Minor Victim-3 with a schoolgirl uniform",
    ),
    DocCfg(id='EFTA00005783', note='heavily redacted handwritten note, 30+ completely redacted pages', date='2019-08-29'),
    DocCfg(id='EFTA01063691', note='inventory of address books and Skype logs seized from Epstein computers'),
    DocCfg(id='EFTA00024275', note='large Wexner funded payments to OB-GYN'),
    DocCfg(id='EFTA00728783', note='list of names and phone numbers'),
    DocCfg(id='EFTA00298379', note='RSVP list for Yom Kippur dinner', date='2010-09-18'),
    DocCfg(id='EFTA01611898', note=f"screenshot of recent contacts in an iPhone"),
    # Urramoor
    DocCfg(
        id='EFTA01107738',
        note=f"creation of {CANTOR_URRAMOOR} with Mr. T's (Prince Andrew?) Urramoor and {LUTNICKS_CANTOR}",
        is_interesting=10,
    ),
    DocCfg(
        id='EFTA01141453',
        note=f"referral agreement between Mr. T's (Prince Andrew?) Urramoor and {LUTNICKS_CANTOR}",
        is_interesting=10,
    ),
    # Replacement text
    DocCfg(
        id='EFTA00004477',
        is_interesting=True,
        display_text='Epstein 50th birthday photo book 12 "THAIS, MOSCOW GIRLS, AFRICA, HAWAII, [REDACTED] [REDACTED], Zorro, [REDACTED] [REDACTED] [REDACTED], CRACK WHOLE PROPOSAL, BALI/THAILAND/ASIA, RUSSIA, [REDACTED], [REDACTED], NUDES, YOGAL GIRLS',
    ),
    DocCfg(id='EFTA00001487', display_text="architectural drawing of one floor of Epstein's NYC residence"),
    DocCfg(id='EFTA00005386', display_text='heavily redacted photo album, lot of photos of girls'),
    DocCfg(id='EFTA01628970', display_text='redacted pictures of naked women'),
    DocCfg(id='EFTA00004070', display_text="photos of Epstein with handwritten caption that didn't OCR well"),
    DocCfg(id='EFTA02731260', display_text='notebook full of handwritten love letters with terrible OCR text'),
    DocCfg(id='EFTA00006100', display_text='Palm Beach Police fax machine activity log 2005-12-28 to 2006-01-04'),
    # Attachments
    DocCfg(id='EFTA00590685', attached_to_email_id='EFTA00830911'),
    # Dates
    DocCfg(id='EFTA02025218', date='2011-09-09'),
    DocCfg(id='EFTA00007693', note="driving directions from Epstein's house to <REDACTED>, flight details for Sandy Berger"),
    # Misc
    fedex_invoice('EFTA00217072', '2005-06-20'),
    fedex_invoice('EFTA00217080', '2005-06-27'),
    imessage_screenshot(id='033434', author=BRAD_EDWARDS, author_uncertain=f"labeled 'Edwards'", is_interesting=False),
    important_messages_pad('EFTA01719859', '2005-10-03'),
    important_messages_pad('EFTA01682477', '2005-04-01'),
    skype_log('032210', recipients=['linkspirit'], is_interesting=True),
    skype_log('018224', recipients=['linkspirit', LAWRENCE_KRAUSS], is_interesting=True),  # we don't know who linkspirit is yet
    skype_log('EFTA01617727'),
    skype_log(
        id='EFTA00507334',
        recipients=[
            'Aleksandra Eriksson',
            ANASTASIYA_SIROOCHENKO,
            'Catherine',
            DANIEL_SIAD,
            DAVID_STERN,
            EMAD_HANNA,
            'Jade Huang',
            NADIA_MARCINKO,
            'sexysearch2010',
            'sophiembh'
        ],
    ),
    skype_log('EFTA01613143', author=MELANIE_WALKER, date='2017-06-24'),
]



import re
from copy import copy
from dataclasses import dataclass, field
from os import devnull
from pathlib import Path
from typing import Literal, Mapping

from rich.align import AlignMethod
from rich.console import Console, RenderableType
from rich.panel import Panel
from rich.padding import PaddingDimensions
from rich.style import Style
from rich.text import Text

from epstein_files.output.html.positioned_rich import PositionedRich, unpack_dimensions, CssProps, OptionalCssProps
from epstein_files.output.rich import CONSOLE_KWARGS, RICH_THEME
from epstein_files.util.constant.html import FONT_FAMILY, HTML_TERMINAL_THEME
from epstein_files.util.helpers.data_helpers import listify, sort_dict_by_keys
from epstein_files.util.helpers.string_helper import quote
from epstein_files.util.logging import logger

# Types
HtmlListTag = Literal['ol', 'ul']
HtmlElements = str | list[str]

# HTML templating vars
CODE_TEMPLATE = '{code}'
SPLITTER = '-# JUNK #-'
SPLITTER_TEMPLATE = SPLITTER + """{stylesheet} {background} {foreground}"""  # these template vars allow export_html() to work
# rich.Console for converting rich renderables to HTML
HTML_CONSOLE_KWARGS = copy(CONSOLE_KWARGS)
HTML_CONSOLE_KWARGS.update({'file': open(devnull, "wt"), 'record': True})

# CSS dicts
CODE_TAG_CSS = {'font-family': 'inherit'}
FONT_CSS_PROPS = {'font-family': FONT_FAMILY}
WIDTH_PROPS = ['max-width', 'width']

# Used by the per object rich -> html renders, e.g print_renderable, panel_to_div, etc.
HTML_RENDER_CONSOLE = Console(**HTML_CONSOLE_KWARGS)


def div_tag(contents: HtmlElements, css_props: OptionalCssProps = None, **kwargs) -> str:
    return tag('div', contents, css_props, **kwargs)


def div_class(contents: HtmlElements, class_name: str, css_props: OptionalCssProps = None, **kwargs) -> str:
    return div_tag(contents, css_props, class_name=class_name, **kwargs)


def div_with_legend(contents: HtmlElements, legend: str, css_props: OptionalCssProps = None, **kwargs) -> str:
    """Use <fieldset> + <legend>: https://css-tricks.com/how-to-add-text-in-borders-using-basic-html-elements/"""
    css_props = css_props or {}

    if legend:
        legend_css_props = {'color': css_props.get('border-color', 'white')}
        contents = tag('legend', legend, legend_css_props) + contents

    return tag('fieldset', contents, css_props, **kwargs)


def from_em(units: str | None) -> int | None:
    if not units:
        return None

    return int(units.removesuffix('em'))


def build_html_list(elements: list[str], list_type: HtmlListTag = 'ul', **kwargs) -> str:
    txt_htmls = [tag('li', t) for t in elements]

    if not elements:
        logger.warning(f"No elements to make <{list_type}> for...")
        return ''

    return tag(list_type, '\n'.join(txt_htmls), **kwargs)


def strip_outer_tag(_html: str, tag: str) -> str:
    html = re.sub(fr"^\s*<{tag}.*?>", '', _html)
    html = re.sub(fr"</{tag}>\s*$", '', html)

    if html == _html:
        logger.warning(f"asked to strip <{tag}> from {html} but nothing stripped...")

    return html


def tag(tag: str, contents: HtmlElements, css_props: OptionalCssProps = None, **kwargs) -> str:
    """Surround `contents` with <tag style="[CSS_STRING]">."""
    tag_kwargs = f' style="{to_inline_css(css_props)}"' if css_props else ''

    if kwargs:
        if 'class_name' in kwargs:
            kwargs['class'] = kwargs.pop('class_name')

        tag_kwargs += ' ' + ' '.join([f'{k}={quote(v)}' for k, v in kwargs.items()])

    return f"<{tag}{tag_kwargs}>{_html_elements_to_str(contents)}</{tag}>"


def to_inline_css(d: Mapping[str, str | int]) -> str:
    return '; '.join([f'{k}: {v}' for k, v in sort_dict_by_keys(d).items() if v])


def _html_elements_to_str(contents: HtmlElements) -> str:
    """no impact on the actual HTML rendering, newlines just for debugging."""
    return '\n'.join(listify(contents))


# "PRE" because it makes <pre> tags
PRE_CONSOLE_TEMPLATE_PREFIX = tag('pre', CODE_TEMPLATE, {})
PRE_CONSOLE_TEMPLATE = PRE_CONSOLE_TEMPLATE_PREFIX + SPLITTER_TEMPLATE



from dataclasses import dataclass, field
import os
from typing import Literal, Self

from rich.align import Align, AlignMethod
from rich.console import RenderableType
from rich.padding import Padding, PaddingDimensions
from rich.table import Table
from rich.text import Text

from epstein_files.util.logging import logger

CssUnit = int | str
CssProps = dict[str, str]
OptionalCssProps = CssProps | None
SideProp = Literal['margin', 'padding']

to_px = lambda pixels: f"{pixels}px" if isinstance(pixels, int) else pixels

# MAKEUP_PADDING to make HTML panel padding match ANSI (there's ~0.5 spaces between "a" and "|" in "a | b" in ansi)
MAKEUP_PADDING = 0.7
BORDER_HORIZONTAL_PADDING = 1
BORDER_VERTICAL_PADDING = MAKEUP_PADDING

# Side constants
Side = Literal['top', 'left', 'right', 'bottom']
HORIZONTAL_SIDES: list[Side] = ['left', 'right']
VERTICAL_SIDES: list[Side] = ['top', 'bottom']
ALL_SIDES: list[Side] = ['top', 'right', 'bottom', 'left']  # Order matters for converting PaddingDimension!

# dimensions are assumed to always be tuple[int, int, int, int] (but as a list)
horizontal_only = lambda dimensions: [0, dimensions[1], 0, dimensions[3]]
vertical_only = lambda  dimensions: [dimensions[0], 0, dimensions[2], 0]

# Margin CSS
dimensions_to_margin_css = lambda dimensions: _dimensions_to_layout_css('margin', dimensions)
margin_horizontal_css = lambda ems: dimensions_to_margin_css((0, ems))
margin_vertical_css = lambda ems: dimensions_to_margin_css((ems, 0))

# Padding CSS
dimensions_to_padding_css = lambda dimensions: _dimensions_to_layout_css('padding', dimensions)
padding_horizontal_css = lambda ems: dimensions_to_padding_css((0, ems))
padding_vertical_css = lambda ems: dimensions_to_padding_css((ems, 0))

# CSS constants
LEFT_JUSTIFIED = {'margin-right': 'auto'}
RIGHT_JUSTIFIED = {'margin-left': 'auto'}
CENTERED = {**LEFT_JUSTIFIED, **RIGHT_JUSTIFIED}
VERTICAL_MARGIN = 1.9

# CSS classes
BLACK_BACKGROUND = 'black_background'
NO_EXPAND = 'no_expand'
BLACK_BG__NO_EXPAND = f"{BLACK_BACKGROUND} {NO_EXPAND}"
PANEL_BODY_CSS_CLASS = f'{NO_EXPAND} document_body_container'


@dataclass
class PositionedRich:
    obj: RenderableType
    align: AlignMethod | None = None
    margin: list[int | float] = field(default_factory=lambda: [0] * 4)
    padding: list[int | float] = field(default_factory=lambda: [0] * 4)  # TODO: currently unused

    @classmethod
    def from_unwrapped_obj(cls, obj: RenderableType, positioned: Self | None = None) -> Self:
        """Alternate constructor to unwrap `Align` and `Padding` and collect the align/padding props in a PositionedRenderable."""
        positioned = positioned or cls(obj)

        if isinstance(obj, Align):
            positioned.align = obj.align
            return cls.from_unwrapped_obj(obj.renderable, positioned)
        elif isinstance(obj, Padding):
            positioned.margin = [obj.top, obj.right, obj.bottom, obj.left]
            return cls.from_unwrapped_obj(obj.renderable, positioned)
        else:
            positioned.obj = obj  # End recursion and set self.obj if obj is not an Align or Padding
            return positioned

    @property
    def alignment_css(self) -> CssProps:
        """margin-auto"""
        return alignment_css(self.align) if self.align else {}

    @property
    def css(self) -> CssProps:
        """Margin and alignment."""
        return {**self.alignment_css, **self.margin_css}

    @property
    def margin_css(self) -> CssProps:
        return dimensions_to_margin_css(self.margin)

    @property
    def margin_horizontal(self) -> list[int | float]:
        return horizontal_only(self.margin)

    @property
    def margin_horizontal_css(self) -> CssProps:
        return dimensions_to_margin_css(self.margin_horizontal)

    @property
    def margins_vertical(self) -> list[int | float]:
        return vertical_only(self.margin)

    @property
    def margins_vertical_css(self) -> CssProps:
        return dimensions_to_margin_css(self.margins_vertical)

    @property
    def margin_top(self) -> int | float:
        return self.margin[0]
    @property
    def margin_right(self) -> int | float:
        return self.margin[1]
    @property
    def margin_bottom(self) -> int | float:
        return self.margin[2]
    @property
    def margin_left(self) -> int | float:
        return self.margin[3]

    @property
    def requires_container_div(self) -> bool | None:
        """
        if both horizontal margin and alignment are set on the same side we need to place the div in a container
        with horizontal padding so the aligned element doesn't go all the way to edge of screen.
        """
        for margin_prop in self.margin_css.keys():
            if self.alignment_css.get(margin_prop):
                logger.info(f"{margin_prop} is set by both self.align='{self.align}' and self.margin={self.margin}, margin value wins out!")
                return True

    def to_html(self) -> str:
        from epstein_files.output.html.builder import table_to_html
        from epstein_files.output.html.elements import div_class, div_tag

        if not isinstance(self.obj, Table):
            raise NotImplementedError(f"Only Table objects can have to_html() called on them")

        # If we need both horizontal margin and horizontal alignment at same time we need two divs:
        if self.requires_container_div:
            #    [INNER DIV] uses only HORIZONTAL ALIGNMENT + vertical MARGIN
            inner_css = {**self.alignment_css, **self.margins_vertical_css}
            inner_div = table_to_html(self.obj, inner_css)

            #    [OUTER DIV] uses only HORIZONTAL MARGIN (as paddinging) with no vertical margin or padding
            outer_div = div_class(inner_div, BLACK_BG__NO_EXPAND, self.margin_horizontal_css)
        else:
            # logger.warning(f"Table does not require inner + outer container to handle conflicting props: {self.css}")
            # TODO: seems liek the BLACK_BG__NO_EXPAND class should be applied somehow?
            return table_to_html(self.obj, self.css)

        return outer_div

    @classmethod
    def zero_dimensions(cls) -> list[int | float]:
        return [0, 0, 0, 0]


# TODO: should this also set text-align?
def alignment_css(align: AlignMethod) -> CssProps:
    """CSS margin-auto props to align things left, right, or center."""
    if align == 'center':
        return dict(CENTERED)
    elif align == 'left':
        return dict(LEFT_JUSTIFIED)
    elif align == 'right':
        return dict(RIGHT_JUSTIFIED)
    elif not align:
        return {}
    else:
        raise ValueError(f"unknown alignment value: '{align}'")


# CSS units
def to_em(num_chars: int | float) -> str:
    """Convert int to CSS em units (1em = size of character in the current font). 1 => "1em"."""
    if isinstance(num_chars, str) and num_chars.endswith('em'):
        logger.error(f"Trying to append a second 'em' to '{num_chars}'")
        return num_chars
    elif num_chars:
        return f"{num_chars}em"
    else:
        return ''


def unpack_dimensions(dimensions: PaddingDimensions) -> list[int]:
    """Unpack an int or 2-tuple to a 4-tuple (top, right, bottom, left)."""
    return list(Padding.unpack(dimensions))


def vertical_spacer(em_units: int | float) -> str:
    return f'<div style="height: {to_em(em_units)}"></div>'


def _dimensions_to_layout_css(prop: SideProp, dimensions: PaddingDimensions) -> CssProps:
    """Turn all non-zero side dimensions into appropriate margin- or padding- CSS props."""
    return {
        f"{prop}-{ALL_SIDES[i]}": to_em(value)
        for i, value in enumerate(unpack_dimensions(dimensions))
        if value
    }



from dataclasses import dataclass, field
from pathlib import Path
from typing import Sequence

from rich import box
from rich.align import Align, AlignMethod
from rich.console import Group, NewLine, RenderableType
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from epstein_files.documents.communication import Communication
from epstein_files.documents.document import Document
from epstein_files.documents.documents.categories import sort_categories
from epstein_files.documents.documents.doc_types_mixin import DocTypesMixin
from epstein_files.documents.email import Email
from epstein_files.documents.emails.emailers import ENTITY_CATEGORIES, get_entities
from epstein_files.documents.messenger_log import MessengerLog
from epstein_files.documents.other_file import OtherFile
from epstein_files.output.layout_elements.file_display import BasePanel, FileDisplay, ListPanel
from epstein_files.output.html.builder import (console_buffer_to_html, render_at_obj_width, panel_to_div,
     render_to_html, text_to_div, write_templated_html)
from epstein_files.output.html.elements import div_class, tag
from epstein_files.output.html.positioned_rich import PositionedRich, to_em, unpack_dimensions, vertical_spacer
from epstein_files.output.rich import console, section_subtitle_panel
from epstein_files.output.site.sites import Site
from epstein_files.people.entity import Entity
from epstein_files.util.env import args, site_config
from epstein_files.util.helpers.data_helpers import listify, uniq_sorted, without_falsey
from epstein_files.util.helpers.rich_helpers import vertically_pad
from epstein_files.util.helpers.string_helper import quote
from epstein_files.util.logging import logger
from epstein_files.util.timer import Timer
from epstein_files.output.title_page import color_key, title_page_top_elements, title_page_bottom_elements

OTHER_FILES_TABLE_MSG = Text("(non emails will appear in tables)", 'gray27 italic')
EMPTY_LINE_HEIGHT = to_em(1.5)
EMPTY_LINE_DIV = f'<div style="height: {EMPTY_LINE_HEIGHT}"></div>'
STICKY_BIO_CSS_CLASS = 'sticky_person_bio_panel'

PrintableObj = Document | FileDisplay
PeopleBiosArg = list[str] | list[Entity] | PrintableObj


@dataclass(kw_only=True)
class DocPrinter(DocTypesMixin):
    """
    Handles printing collections of documents with biographical info panels interspersed.

    Args:
        epstein_files (EpsteinFiles): the data
        collect_other_files_to_tables (bool, optional): OtherFiles will be collected into small tables if they are sequential
        html_elements (list[str]): HTML for all objects printed so far
        printed_docs (list[Document]): all Documents that have been printed so far
        printed_name_bios (set[Entity]): all the names for which biographical information has been printed already
    """
    epstein_files: 'EpsteinFiles'
    collect_other_files_to_tables: bool = True
    html_elements: list[str] = field(default_factory=list)
    printed_entity_bios: set[Entity] = field(default_factory=set)
    _last_bio_panel = ''
    _other_files_queue: list[OtherFile] = field(default_factory=list)
    _suppressed_docs_queue: list[Document] = field(default_factory=list)

    def __post_init__(self):
        """Initialize with HTML of what's been printed to the console history buffer up to this point."""
        self.html_elements.append(console_buffer_to_html(console, False))

    @property
    def printed_docs(self) -> list[Document]:
        return Document.filter_for_type(self._documents)  # TODO: necessary because of FileDisplay objs

    @property
    def printed_emails(self) -> list[Email]:
        return self.emails

    @property
    def printed_ids(self) -> list[str]:
        return [f.file_id for f in self.printed_docs]

    def line(self, num: int = 1) -> None:
        """Print blank lines to HTML and terminal, similar to `console.line()`."""
        self.html_elements.append(vertical_spacer(num))
        console.line(num)

    def new_entities(self, names_or_doc: PeopleBiosArg) -> list[Entity]:
        """List of names found in relation to `names_or_doc` that have not been biographically printed before."""
        if isinstance(names_or_doc, FileDisplay) or not names_or_doc:
            return []
        elif isinstance(names_or_doc, Document):
            entities = names_or_doc.entity_scan(exclude=list(self.printed_entity_bios))
        else:
            entities = get_entities(names_or_doc)

        return [e for e in entities if e not in self.printed_entity_bios]

    def new_entities_with_bios(self, names_or_doc: PeopleBiosArg) -> list[Entity]:
        """Return names that a) were not seen before and b) have configured biographical info."""
        return [n for n in self.new_entities(names_or_doc) if n.has_bio]

    def print_biographies(self) -> None:
        """Print all the `Entity` objects that are configured in one big colored list."""
        self.print_section_subtitle('Entities With Configured Biographical Info')

        entity_bios = [
            Padding(c.bio_txt, site_config.contact_list_padding)
            for category in sort_categories([c for c in ENTITY_CATEGORIES.keys()])
            for c in ENTITY_CATEGORIES[category]
        ]

        self.print(entity_bios)

    def print_centered(self, _renderables: RenderableType | list[RenderableType]):
        renderables = listify(_renderables)

        if any(isinstance(r, Align) for r in renderables):
            logger.warning(f"wrapping an Align in another Align...")

        self.print([Align.center(obj) for obj in listify(renderables)])

    def print_color_key(self) -> None:
        self.print_centered(color_key())

    def print_documents(
        self,
        docs: Sequence[PrintableObj],
        suppressed_as_normal: bool = False,
        log_sfx: str = '',
    ) -> None:
        """
        Sequential suppression msgs + OtherFiles collect in queues to be printed when
        an obj of another type shows up or if there's new names for a biography panel.

        Args:
            docs (Sequence[PrintableObj]): objs to print
            suppressed_as_normal (bool, optional): if True docs with suppression msgs will be treated like normal Documents
            log_sfx (str, optional): just for log messages
        """
        suppressed_docs: list[Document] = []
        process_suppressed_docs_queue = lambda: suppressed_docs.extend(self._print_suppression_msgs_queue())
        timer = Timer()

        if (should_log_in_intervals := (len(docs) > 1000)):
            logger.info(f"{type(self).__name__}.print_documents() called with {len(docs):,} objects {log_sfx}")

        for i, doc in enumerate(docs, 1):
            # Handle sequences of uninteresting or otherwise suppressed docs
            if isinstance(doc, Document) and doc.suppressed_txt and not suppressed_as_normal:
                self._log_state(doc, f"suppressing {quote(doc.suppressed_txt.plain)}")
                self._suppressed_docs_queue.append(doc)
                continue

            process_suppressed_docs_queue()

            # Collect sequences of otherFile objects into a table
            if self.collect_other_files_to_tables and isinstance(doc, OtherFile) and doc.is_valid_for_table:
                if (new_entities := self.new_entities_with_bios(doc)):
                    doc._log(f"Caching biographic panel for new entities: {[str(e) for e in new_entities]}")
                    self._cache_biographies_panel(new_entities)

                self._log_state(doc, f"queueing other file for table")
                self._other_files_queue.append(doc)
                continue

            self._log_state(doc, f"triggering print of other files queue")
            self._print_other_files_queue()
            self.print(doc)

            if should_log_in_intervals and (i % 100 == 0):
                timer.print_at_checkpoint(f"Printed {i:,} objs of {len(docs):,} ({len(suppressed_docs):,} suppressed) {log_sfx}")

        process_suppressed_docs_queue()
        self._print_other_files_queue()
        timer.print_at_checkpoint(f"Finished printing {len(docs):,} objs ({len(suppressed_docs):,} suppressed) {log_sfx}")

    def print(self, renderables: RenderableType | Sequence[RenderableType]) -> None:
        """All things being printed should come through here, which collects both terminal and HTML output as its written."""
        for renderable in listify(renderables):
            positioned = PositionedRich.from_unwrapped_obj(renderable)

            # Add html string for this renderable to self.html_elements
            if isinstance(positioned.obj, NewLine):
                self.line()
                continue
            elif isinstance(positioned.obj, PrintableObj):
                doc_bios_html = self._build_biographies_panel_html(self.new_entities_with_bios(positioned.obj))
                self._append_element_with_bio_div(positioned.obj.to_html(), doc_bios_html)
                doc = positioned.obj.document if isinstance(positioned.obj, FileDisplay) else positioned.obj
                self._documents.append(doc)
            elif isinstance(positioned.obj, Table):
                html_table = positioned.to_html()  # TODO: currently the only type that delegates to the PositionedRich obj to get HTML

                # this is a bad way to signify the table being printed is an OtherFiles table
                if self._last_bio_panel:
                    self._append_element_with_bio_div(html_table)
                else:
                    self.html_elements.append(html_table)
            elif isinstance(positioned.obj, Panel):
                self.html_elements.append(panel_to_div(positioned.obj, positioned.css))
            elif isinstance(positioned.obj, BasePanel):
                margin = unpack_dimensions((site_config.indents.info, 0))  # TODO: this margin dimension should only exist on eone side if aligned
                self.html_elements.append(positioned.obj.to_div(margin))
            elif isinstance(positioned.obj, Text):
                self.html_elements.append(text_to_div(positioned.obj, positioned.css))
            elif '__rich__' in dir(positioned.obj) and (rich_obj := positioned.obj.__rich__()):
                if isinstance(rich_obj, Text):
                    element_html = text_to_div(rich_obj, positioned.css)
                else:
                    logger.warning(f"printinng possibly not fully supported object type {type(positioned.obj).__name__}\n{positioned.obj}")
                    element_html = render_to_html(rich_obj)

                self.html_elements.append(element_html)
            elif isinstance(positioned.obj, str):
                self.html_elements.append(tag('p', positioned.obj, positioned.css))
            else:
                raise TypeError(f"renderable of unsupported type: {type(positioned.obj).__name__}: {positioned.obj}")

            # TODO: console.print at end after HTML is generated so the bios panel will be cached first
            # TODO: because build_biographies_panel_html() has the side effect of printing to the console.
            console.print(renderable)

    def print_section_subtitle(self, msg: str) -> None:
        self.print_centered(section_subtitle_panel(msg))

    def print_title_page_top(self) -> None:
        self._print_title_page_elements(title_page_top_elements())

    def print_title_page_bottom(self) -> None:
        self._print_title_page_elements(title_page_bottom_elements(self.epstein_files))

    def write_html(self, write_to: Path | Site) -> Path:
        """Export custom HTML, trigger rich export_html() if `SiteType` given, returns custom HTML path."""
        if isinstance(write_to, Path):
            output_path = write_to
            logger.warning(f"Not exporting rich.console to HTML directly (only custom)...")
        else:
            output_path = Site.custom_html_build_path(write_to)
            from epstein_files.output.output import write_html
            write_html(write_to)

        return write_templated_html(self.html_elements, output_path)

    def _align_biographical_panel(self, panel: Panel) -> Align:
        return Align(Padding(panel, site_config.character_bio_padding), 'right')

    def _append_element_with_bio_div(self, element: str, bio_panel: str = '') -> None:
        """Cache the `last_bio_panel` html so it can be placed inside div with document."""
        if self._last_bio_panel:
            bio_panel = self._last_bio_panel
            self._last_bio_panel = ''

        element = '\n'.join([bio_panel, element])
        self.html_elements.append(div_class(element, 'doc_container'))

    def _biographical_panel(self, _entities: list[Entity]) -> Panel | None:
        """Panel showing biographical info for a list of names."""
        if (entities := self.new_entities_with_bios(_entities)):
            bios = [Text('', justify='right').append(e.bio_txt) for e in entities if e.has_bio]
            logger.debug(f"Creating bios panel for {len(entities)} new names (out of {len(_entities)} _entities)")

            return Panel(
                Group(*bios),
                border_style='dim',
                box=box.DOUBLE,
                expand=False,
                style='on gray7',
                title=Text(f"new names in next file", 'grey85 italic'),
                title_align='right',
            )
        else:
            return None

    def _build_biographies_panel_html(self, entities: list[Entity]) -> str:
        """NOTE: Also prints to console!!"""
        if (bio_panel := self._biographical_panel(entities)):
            self._print_other_files_queue()  # Clear the queue before new biographical panel
            console.print(self._align_biographical_panel(bio_panel))  # TODO: has side effect of printing to console
            panel_html = render_at_obj_width(bio_panel)
            self.printed_entity_bios.update(entities)
            return div_class(panel_html, STICKY_BIO_CSS_CLASS)
        else:
            return ''

    def _cache_biographies_panel(self, entities: list[Entity]) -> None:
        """The html string of the bios panel is cached in `self.last_bio_panel`."""
        if (bios_html := self._build_biographies_panel_html(entities)):
            if self._last_bio_panel:
                raise RuntimeError(f"last_bio_panel should be empty, instead it's:\n{self._last_bio_panel}")
            else:
                self._last_bio_panel = bios_html

    def _log_state(self, doc: PrintableObj, msg: str = '') -> None:
        supressed_ids = [f.file_id for f in self._suppressed_docs_queue]
        other_files_queue_ids = [f.file_id for f in self._other_files_queue]
        queues_str = f"(suppressed queue: {supressed_ids}, other files queue: {other_files_queue_ids})"

        if isinstance(doc, FileDisplay):
            logger.info(f"{doc} {msg} {queues_str}")
        else:
            doc._log(f"{msg} {queues_str}")

    def _print_other_files_queue(self) -> None:
        """Print any queued OtherFile objects collected in a table."""
        if not self._other_files_queue:
            return

        # Title is only on the first OtherFiles table printed
        if any(isinstance(d, OtherFile) for d in self.printed_docs):
            table_title = None
        else:
            table_title = OTHER_FILES_TABLE_MSG
            self.line()

        table = OtherFile.files_preview_table(self._other_files_queue, title=table_title, title_justify='center')
        self.print(Padding(table, (0, 0, 1, site_config.indents.other_files_table)))
        self._documents.extend(self._other_files_queue)
        self._other_files_queue = []

    def _print_suppression_msgs_queue(self) -> list[Document]:
        """Print any suppression messages. Returns Documents whose suppression msg was printed."""
        if not self._suppressed_docs_queue:
            return []

        suppressed_txts = without_falsey([d.suppressed_txt for d in self._suppressed_docs_queue])
        msgs_panel = ListPanel(border_style='', text=suppressed_txts)
        self.print(msgs_panel)
        console.line()  # TODO this isn't happening in HTML output

        processed_suppressed_docs = self._suppressed_docs_queue
        self._suppressed_docs_queue = []  # Reset queue
        return processed_suppressed_docs

    def _print_title_page_elements(self, renderables: list[RenderableType]) -> None:
        """"Centered and vertically paded Tables and Panels."""
        renderables = [vertically_pad(r) if isinstance(r, (Panel, Table)) else r for r in renderables]
        self.print_centered(renderables)



import re
from datetime import datetime

from epstein_files.documents.document import Document
from epstein_files.documents.email import Email
from epstein_files.output.rich import console
from epstein_files.people.names import ADA_CLAPP, CHRISTOPHER_DILORIO, JEFFREY_EPSTEIN, LEON_BLACK, MELANIE_SPINELLA, sort_names
from epstein_files.util.constant.strings import LEON_BLACK_EMAIL_ID
from epstein_files.util.env import args
from epstein_files.util.helpers.data_helpers import groupby
from epstein_files.util.logging import logger

DILORIO_SPLIT = '\nFrom: Chris'
LEON_BLACK_EMAIL_REGEX = re.compile(r"^(From: .{,50}\nDate:|Date: ).*?(?=(From|Date|\Z))", re.DOTALL | re.MULTILINE)
LEON_BLACK_FWD_REGEX = re.compile(r"^-+(Forwarded|Original) message-+", re.MULTILINE)
LEON_BLACK_OCR_REPAIRS = {re.compile('^ate: ', re.MULTILINE): 'Date: '}
TO_JEFFREY_REGEX = re.compile(r"^Jeffrey-", re.MULTILINE)
TO_LEON_REGEX = re.compile(r"^Leon,", re.MULTILINE)


def split_up_multi_email_files(big_emails: list[Email]) -> list[Email]:
    """
    Some files have 100+ emails concatenated into one file so we split them up in a bespoke way.
    Has side effect of setting `_was_split_up=True` for all `big_emails` which removes them from
    `EpsteinFiles.documents` list so they can only be accessed through `EpsteinFiles._documents`.
    """
    author_emails = groupby(big_emails, lambda e: e.author)
    authors = sort_names([name for name in author_emails.keys()])
    assert authors == [None, CHRISTOPHER_DILORIO], f"Unexpected author in big_emails: {authors}"

    for big_email in big_emails:
        big_email._was_split_up = True

    return [
        *_split_up_dilorio_whistleblower_emails(author_emails[CHRISTOPHER_DILORIO]),
        *_split_up_leon_black(author_emails[None]),
    ]


def _split_up_dilorio_whistleblower_emails(dilorio_emails: list[Email]) -> list[Email]:
    """Dilorio is some kind of finance whistleblower. his emails are not Epstein related but very interesting."""
    sub_emails = []
    skipped = []

    for big_email in dilorio_emails:
        texts = [
            (f"{DILORIO_SPLIT}{t}" if i > 0 else t).strip()
            for i, t in enumerate(big_email.text.split(DILORIO_SPLIT))
        ]

        logger.warning(f"Parsing {big_email} into {len(texts)} sub emails...")

        for i, text in enumerate(texts, 1):
            new_file_stem = _new_file_stem(big_email, i)
            email = Email(big_email.file_path.parent.joinpath(new_file_stem), text=text)
            email.extracted_author = big_email.author
            email.extracted_recipients = [] if email.extracted_recipients == ['Rnignc Lapnai PP'] else email.extracted_recipients
            email.extracted_recipients = email.extracted_recipients or ['SEC']
            email.header.to = email.header.to or ['SEC']  # avoids email being considered as having unknown recipient

            if not email.actual_text or email.lines[-1].startswith('Subject:'):
                email._warn(f"skipping empty {email.author} email...")
                skipped.append(email)
                continue

            sub_emails.append(email)

    return _uniquify_by_timestamp(sub_emails, dilorio_emails, 'dilorio', len(skipped))


def _split_up_leon_black(big_emails: list[Email]) -> list[Email]:
    """Create emails from gigantic Leon Black email."""
    assert len(big_emails) == 1, f"have {len(big_emails)} Leon Black emails, should have 1"
    big_email = big_emails[0]
    assert big_email.file_id == LEON_BLACK_EMAIL_ID, f"wrong Leon Black email {big_email.file_id}"
    big_text = big_email.repair_ocr_text(LEON_BLACK_OCR_REPAIRS, big_email.text)
    emails: list[Email] = []
    skipped = []
    fwded_email_text = ''

    for i, match in enumerate(LEON_BLACK_EMAIL_REGEX.finditer(big_text), 1):
        text = match.group(0)
        new_file_stem = _new_file_stem(big_email, i)

        if LEON_BLACK_FWD_REGEX.search(text):
            logger.warning(f"Fwded email detected in Leon Black email body, appending next email...")
            fwded_email_text = fwded_email_text + text
            continue
        else:
            email_text = fwded_email_text + text
            fwded_email_text = ''

        email = Email(big_email.file_path.parent.joinpath(new_file_stem), text=email_text)
        text_to_scan_for_recipients = email.actual_text[:120]

        if (TO_LEON_REGEX.search(email.text) and LEON_BLACK not in email.extracted_recipients) \
                or email.extracted_recipients == [MELANIE_SPINELLA]:
            email.extracted_recipients = sort_names(email.extracted_recipients + [LEON_BLACK])

        if 'Clapp' in text_to_scan_for_recipients:
            email.extracted_recipients += [ADA_CLAPP]

        if 'Avantario' in text_to_scan_for_recipients:
            email.extracted_recipients += ['Joe Avantario']

        if email.recipients and not (email.extracted_author or TO_JEFFREY_REGEX.search(text_to_scan_for_recipients)):
            email.extracted_author = JEFFREY_EPSTEIN

            if JEFFREY_EPSTEIN in email.extracted_recipients:
                email.extracted_recipients = [r for r in email.extracted_recipients if r != JEFFREY_EPSTEIN]

        if args.raw or args.debug:
            print(f"\n--- RAW TEXT {new_file_stem} ---\n{email_text}\n--- END RAW TEXT {new_file_stem} ---\n")
            console.print(email)
            console.print(email._debug_txt())

        emails.append(email)

    return _uniquify_by_timestamp(emails, [big_email], LEON_BLACK, len(skipped))


def _new_file_stem(email: Email, i: int) -> str:
    return email.file_info.file_stem + f'_{i}.txt'


def _uniquify_by_timestamp(emails: list[Email], from_emails: list[Email], label: str, num_skipped: int = 0) -> list[Email]:
    """Uniquify by timestamp to eliminate dupes."""
    # Validate timestamps
    for email in emails:
        if email.timestamp is None or email.timestamp > Email.MAX_TIMESTAMP:
            raise ValueError(f"{email} bad timestamp in extracted email: {email.timestamp}")

    email_at_timestamp = {e.timestamp: e for e in Document.sort_by_id(emails)}

    logger.warning(
        f"Created {len(emails)} ({len(email_at_timestamp)} unique) Emails from {len(from_emails)} {label} emails" \
        f", skipped {num_skipped} empty sub emails"
    )

    return [e for e in email_at_timestamp.values()]



"""
Mobile vs. non-mobile.
"""
from dataclasses import dataclass
from datetime import datetime
from typing import ClassVar

from rich.padding import PaddingDimensions
from rich.text import Text

from epstein_files.util.constant.strings import SUBHEADER_STYLE, TIMESTAMP_STYLE
from epstein_files.util.external_link import link_markup
from epstein_files.util.helpers.string_helper import capitalize_first, iso_timestamp, starred_header, timestamp_human

ALL_OTHER_FILES_MULTIPLIER = 1.5  # applied when --all-other-files is in use
DEFAULT_WIDTH = 150
NOT_ALL_FILES_MSG = 'Not All Epstein Files Are Here!'

RICH_LINK_MARKUP = link_markup('https://pypi.org/project/rich/', 'rich')
MOBILE_WARNING = f"This page was generated by {RICH_LINK_MARKUP}. It is very poorly optimized for mobile and may take a while to render on a phone." \
                  "\nIf your browser won't render part of the page try reloading."
MOBILE_WARNING_TXT = Text.from_markup(MOBILE_WARNING)
MOBILE_WARNING_TXT.stylize('dim')


@dataclass(kw_only=True)
class IndentCfg:
    """
    show_with (int): Indent for `show_with_name` files
    """
    email_attachment: int
    email_body: int
    info: int
    other_files_table: int
    show_with: int
    supressed_msg: int


# TODO: another thing that changes based on mobile is id_only for link texts
@dataclass
class MobileConfig:
    abbreviations_width: ClassVar[int | None] = None
    character_bio_padding: ClassVar[tuple[int, int, int, int]] = (0, 0, 1, 2)
    contact_list_padding: ClassVar[PaddingDimensions] = (0, 0, 0, 1)
    max_alt_links: ClassVar[int | None] = 1
    not_all_files_warning: ClassVar[str] = starred_header(NOT_ALL_FILES_MSG, num_spaces=1, num_stars=1)
    num_color_key_cols: ClassVar[int] = 2
    other_files_preview_chars: ClassVar[int] = 300
    section_header_padding: ClassVar[PaddingDimensions] = (2, 1, 1, 1)
    show_emailer_tables: ClassVar[bool] = False
    site_glossary_horizontal_padding: ClassVar[int] = 2
    social_link_separator: ClassVar[str] = ' '
    subtitle_margins: ClassVar[PaddingDimensions] = (2, 1, 1, 1)
    subtitle_padding: ClassVar[PaddingDimensions] = (0, 1)
    subtitle_width: ClassVar[int | None] = None
    width: ClassVar[int] = 52

    indents: ClassVar[IndentCfg] = IndentCfg(
        email_attachment=6,
        email_body=0,
        info=1,
        other_files_table=0,
        show_with=0,
        supressed_msg=0,
    )

    @classmethod
    def format_text_msg_time(cls, dt: datetime) -> str:
        return dt.strftime(r"%-d/%-m/%-y %-H:%M")

    @classmethod
    def info_padding(cls) -> PaddingDimensions:
        return (0, 0, 0, cls.indents.info)

    @classmethod
    def email_subheader(cls, email_type: str, author: Text, recipients: Text, timestamp: datetime | Text) -> Text:
        prefix = f"{capitalize_first(email_type)} from " if email_type != 'email' else ''
        txt = Text(prefix, SUBHEADER_STYLE).append(author)

        if isinstance(timestamp, Text):
            timestamp = timestamp
        else:
            timestamp = Text(' at ').append(timestamp_human(timestamp), TIMESTAMP_STYLE)

        txt.append(' to ').append(recipients).append(timestamp)
        return txt

    @classmethod
    def suppressed_file_padding(cls) -> tuple[int, int, int, int]:
        return (0, 0, 0, cls.indents.supressed_msg)


class SiteConfig(MobileConfig):
    abbreviations_width: ClassVar[int | None] = 62
    character_bio_padding: ClassVar[tuple[int, int, int, int]] = (0, 1, 0, 0)
    contact_list_padding: ClassVar[PaddingDimensions] = (0, 0, 0, 5)
    max_alt_links: ClassVar[int | None] = None
    not_all_files_warning: ClassVar[str] = starred_header(NOT_ALL_FILES_MSG, num_spaces=6, num_stars=14)
    num_color_key_cols: ClassVar[int] = 5
    other_files_preview_chars: ClassVar[int] = 900
    section_header_padding: ClassVar[PaddingDimensions] = (2, 25, 1, 25)
    show_emailer_tables: ClassVar[bool] = True
    site_glossary_horizontal_padding: ClassVar[int] = 5
    social_link_separator: ClassVar[str] = '  /  '
    subtitle_margins: ClassVar[PaddingDimensions] = (2, 0, 2, 0)
    subtitle_padding: ClassVar[PaddingDimensions] = (0, 3)
    subtitle_width: ClassVar[int | None] = 70
    width: ClassVar[int] = DEFAULT_WIDTH

    indents: ClassVar[IndentCfg] = IndentCfg(
        email_attachment=12,
        email_body=1,
        info=1,
        other_files_table=2,
        show_with=30,
        supressed_msg=1,
    )

    @classmethod
    def format_text_msg_time(cls, dt: datetime) -> str:
        return iso_timestamp(dt)

    @classmethod
    def email_subheader(cls, email_type: str, author: Text, recipients: Text, timestamp: datetime) -> Text:
        return super().email_subheader(
            f"OCR text of {email_type}",
            author,
            recipients,
            Text(f" probably sent at ").append(timestamp_human(timestamp), style=TIMESTAMP_STYLE)
        )



"""
slightly more abstract / higher level functions for turning rich objs to HTML
"""
from copy import deepcopy
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

from rich.align import Align
from rich.panel import Panel
from rich.console import RenderableType
from rich.padding import Padding
from rich.style import Style
from rich.table import Table
from rich.text import Text

from epstein_files.output.html.elements import _html_elements_to_str
from epstein_files.output.html.elements import *
# from epstein_files.output.html.positioned_rich import to_em, to_px, PositionedRich, unpack_dimensions
from epstein_files.output.html.positioned_rich import *
from epstein_files.output.html.rich_style import RichStyle
from epstein_files.output.rich import console
from epstein_files.util.constant.html import CUSTOM_HTML_TEMPLATE, HTML_TERMINAL_THEME
from epstein_files.util.env import args
from epstein_files.util.helpers.data_helpers import add_constant, listify, update_truthy
from epstein_files.util.helpers.file_helper import log_file_write
from epstein_files.util.helpers.string_helper import join_truthy
from epstein_files.util.logging import logger

CSS = Path(__file__).parent.joinpath('page.css').read_text()
BOX_BORDER_RADIUS = 4
BOX_BORDER_WIDTH = 2
MAX_RENDER_WIDTH = 1024
VERTICAL_MARGIN_EMS = to_em(VERTICAL_MARGIN)
BOTTOM_MARGIN_PROPS = {'margin-bottom': VERTICAL_MARGIN_EMS}

# Makes the inset border work
OUTER_PANEL_DIV_PADDING_CSS = dimensions_to_padding_css(
    (
        BORDER_VERTICAL_PADDING,
        BORDER_VERTICAL_PADDING,
    )
)

DEFAULT_HTML_RENDERER_KWARGS = {
    'code_format': PRE_CONSOLE_TEMPLATE,
    'inline_styles': True,
    'theme': HTML_TERMINAL_THEME,
}

# These are the props for the outer panel
PANEL_BASE_PROPS = {
    "display": "inline-block",
    'width': 'fit-content',
    **OUTER_PANEL_DIV_PADDING_CSS,
}

PANEL_BORDER_PROPS = {
    "border-radius": to_px(BOX_BORDER_RADIUS),
    "border-style": "solid",
    "border-width": to_px(BOX_BORDER_WIDTH),
}


# TODO: refactor this into RichStyle
def border_css_props(style: str | Style | None) -> dict[str, str]:
    """CSS props to make an HTML div with border-color set to `style` arg as a standard CSS RGB string."""

    if style and (html_style := RichStyle(style)).foreground_color_hex:
        return {
            "border-color": html_style.foreground_color_hex,
            **PANEL_BORDER_PROPS
        }
    else:
        return {}


def console_buffer_to_html(_console: Console, clear: bool = True) -> str:
    """
    Export the current `console` write buffer to an HTML string.
    SPLITTER is used to strip off cruft that rich export_html() creates because it prints to variables we don't want.
    """
    return _console.export_html(clear=clear, **DEFAULT_HTML_RENDERER_KWARGS).split(SPLITTER)[0]


def one_row_table_html(table: Table, css_props: OptionalCssProps = None) -> str:
    """
    Convert a one row/one column `Table` to HTML string in a way that looks like a normal panel div not a table.
    These kind of `Table`s are currently rendered by `Email` objs (the `config.description` is the "header").
    """
    css_props = css_props or {}
    col1 = table.columns[0]
    header_div = ''

    if table.show_header:
        if isinstance(col1.header, Text):
            header = col1.header
        elif isinstance(col1.header, str):
            header = Text(col1.header, justify=col1.justify)
        else:
            raise ValueError(f"invalid header type {type(col1.header).__name__} {col1.header}")

        header_props = {
            'border-bottom-color': RichStyle(table.border_style).foreground_color_hex,
            'text-align': header.justify or 'left',
            **(RichStyle(table.header_style).to_css if table.header_style else {}),
        }

        header_span = render_to_html(Text('', style=col1.header_style or '').append(header))
        header_div = div_class(header_span, 'document_panel_header', header_props)

    body_html = render_to_html(Text('', style=col1.style).append(col1._cells[0]))
    body_div = div_class(body_html, PANEL_BODY_CSS_CLASS, PANEL_BASE_PROPS)

    logger.debug(f"one_row_table_html(): border_style='{table.border_style}'")
    logger.debug(f"                      css_props: {css_props}")
    logger.debug(f"                      border_css_props: {border_css_props(table.border_style)}")
    logger.debug(f"                      returning div with BLACK_BG__NO_EXPAND\n")

    return div_class(
        join_truthy(header_div, body_div, '\n'),
        BLACK_BG__NO_EXPAND,
        {
            'text-align': col1.justify,
            **border_css_props(table.border_style),
            **css_props
        }
    )


def panel_to_div(panel: Panel, css_props: OptionalCssProps) -> str:
    """
    Convert a rich `Panel` to a div that looks the same.
    this requires two divs with padding between them to approximate the inset border of rich ansi
    css_props are applied to the OUTER div.
    """
    css_props = css_props if css_props is not None else BOTTOM_MARGIN_PROPS                            # TODO: Default bottom margin seems wrong
    border_style = panel.border_style if panel.style == 'none' or not panel.style else panel.style     # TODO: what's up with the string 'none'??
    inner_div_padding_dims = add_constant(unpack_dimensions(panel.padding or 0), MAKEUP_PADDING)
    html_style = RichStyle(panel.style)

    inner_div_css = {
        **html_style.to_css,
        **border_css_props(border_style),
        **dimensions_to_padding_css(inner_div_padding_dims),
    }

    outer_div_css = {
        **html_style.to_css,
        **PANEL_BASE_PROPS,
        **css_props,
    }

    # Set the width but only on the inner div. outer div will adjust automatically based on its padding
    if panel.width:
        for prop in WIDTH_PROPS:
            inner_div_css[prop] = to_em(panel.width - MAKEUP_PADDING)

    logger.debug(f"panel_to_div(): panel.style='{panel.style}', panel.border_style='{panel.border_style}'\n\n   inner_div_css props: {inner_div_css}\n\n    outer_div_css: {outer_div_css}\n")
    inner_div = div_class(render_at_css_width(panel.renderable), 'panel inner_panel', inner_div_css)
    return div_class(inner_div, 'panel outer_panel', outer_div_css)


def render_at_console_width(obj: RenderableType) -> str:
    """render at the width currently used by the main rich.Console obj."""
    return _render_at_width(obj, console.width)


def render_at_css_width(obj: RenderableType, props: OptionalCssProps = None) -> str:
    """Convert CssProps to rich Padding units."""
    props = props or {}
    css_width = from_em(props.get('max-width', props.get('width')))

    if css_width:
        logger.debug(f"render_at_css_width() extracted width of {css_width} from `props` arg CSS:\n   {props}\n")
    else:
        logger.debug(f"no width found in CSS when render_at_css_width() called, no point in using this function.")

    return _render_at_width(obj, css_width or HTML_RENDER_CONSOLE.width)


def render_at_obj_width(obj: RenderableType) -> str:
    """Render `obj` to a <pre> block with same width as `obj`."""
    with _obj_width_renderer(obj) as _renderer:
        return render_to_html(obj, _renderer)


def render_max_width(obj: RenderableType) -> str:
    """render obj to a <pre> block formatted entirely by rich monospace layout that is MAX_RENDER_WIDTH chars wide."""
    return _render_at_width(obj, MAX_RENDER_WIDTH)


def render_to_html(obj: RenderableType, renderer_console: Console | None = None) -> str:
    """Convert rich renderable to HTML <pre> str by printing it then calling `console.export_html()`."""
    renderer_console = renderer_console or HTML_RENDER_CONSOLE
    renderer_console.print(obj, end='')
    return console_buffer_to_html(renderer_console)


def table_to_html(table: Table, css_props: OptionalCssProps = None) -> str:
    """Convert a rich `Table` to an HTML table that looks the same."""
    css_props = css_props or {}
    col_styles = [col.style or '' for col in table.columns]
    header_css_props = RichStyle(table.header_style).to_css if table.header_style else {}

    if table.title:
        title_txt = table.title.copy() if isinstance(table.title, Text) else Text(table.title)

        if table.title_style:
            title_txt.stylize(table.title_style)

        title_justify = 'center' if table.title_justify in ['default', 'full'] else table.title_justify
        title_css = alignment_css(title_justify)

        if title_justify == 'center':
            title_css['text-align'] = 'center'

        title_html = div_class(render_to_html(title_txt), 'table_title', title_css)
    else:
        title_html = ''

    row_props = {'border-bottom': '1px solid dimgray'} if table.show_lines else {}
    cell_props = {**row_props, **(padding_vertical_css(MAKEUP_PADDING) if table.show_lines else {})}

    headers = [
        _table_cell(
            Text('', style=col.header_style or '').append(col.header),
            {
                'max-width': to_em(col.max_width or col.width),
                'text-align': ((col.header.justify or '') if isinstance(col.header, Text) else '') or col.justify,
                **header_css_props
            },
            'columnheader'
        )
        for col in table.columns
    ]

    rows = [
        [
            _table_cell(
                Text('', style=col_styles[j]).append(col._cells[i]) if isinstance(col._cells[i], (str, Text)) else col._cells[i],
                {
                    'max-width': to_em(col.max_width or col.width),
                    'text-align': table.columns[j].justify,
                    **cell_props
                }
            )
            for j, col in enumerate(table.columns)
        ]
        for i, _row in enumerate(table.rows)
    ]

    # table.caption is the text at the bottom under the table, the footer kinda
    if table.caption:
        caption = table.caption if isinstance(table.caption, Text) else Text(table.caption)
        caption_style = RichStyle(table.caption_style or '')

        caption_html = text_to_div(
            caption,
            {
                'margin-top': to_em(0.4),  # TODO: make this config
                'text-align': 'center',
                'width': '90%',
                **css_props,
                **caption_style.to_css,
                **CENTERED
            }
        )
    else:
        caption_html = ''

    rows = [headers, *rows] if table.show_header else rows
    html_rows = [div_class('\n'.join(row), 'row', row_props, role='row') for row in rows]

    # Do positioning and layout
    table_html = div_class(
        html_rows,
        'table',
        border_css_props(table.border_style),
        role='table'
    )

    return div_class(
        _html_elements_to_str([title_html, table_html, caption_html]),
        'table_container',
        {**BOTTOM_MARGIN_PROPS, **css_props},
    )


def text_to_div(txt: Text, css_props: OptionalCssProps, class_name: str = BLACK_BG__NO_EXPAND) -> str:
    """Wrap a line or block of text in a plain black background div."""
    if css_props:
        logger.debug(f"text_to_div() called with css_props arg:\n  {css_props}\n")

    return div_class(render_at_css_width(txt), class_name, css_props)


def text_to_list(textish: list[str] | list[Text], list_type: HtmlListTag = 'ul', **kwargs) -> str:
    """Convert a list of strings or a string with newlines in it to HTML list (<ul> or <ol> tag)."""
    txts = [Text(t) if isinstance(t, str) else t for t in textish]
    html_elements = [strip_outer_tag(render_max_width(txt), 'pre') for txt in txts]
    return build_html_list(html_elements, list_type, **kwargs)


@contextmanager
def tmp_console(tmp_width: int) -> Generator[Console, None, None]:
    """safely use a rich console with temporary width for rendering rich -> HTML."""
    old_console_width = HTML_RENDER_CONSOLE.width
    HTML_RENDER_CONSOLE.width = tmp_width

    try:
        yield HTML_RENDER_CONSOLE
    finally:
        HTML_RENDER_CONSOLE.width = old_console_width


def write_templated_html(elements: list[str] | str, output_path: Path) -> Path:
    """Render a collection of HTML elements to an HTML file. Returns file that was written."""
    body = '\n\n'.join(listify(elements))

    html = str(CUSTOM_HTML_TEMPLATE).format(
        code=body,
        stylesheet=CSS,
        background=HTML_TERMINAL_THEME.background_color.hex,
        foreground=HTML_TERMINAL_THEME.foreground_color.hex,
    )

    output_path.write_text(html)
    log_file_write(output_path)
    return output_path


@contextmanager
def _obj_width_renderer(obj: RenderableType) -> Generator[Console, None, None]:
    """temporarily set html_renderer.width to this object's width to constrain rich output."""
    with tmp_console(HTML_RENDER_CONSOLE.measure(obj).minimum) as _tmp_console:
        yield _tmp_console


def _render_at_width(obj: RenderableType, width: int) -> str:
    """render `obj` to HTML but temporarily change the console width."""
    with tmp_console(width) as _renderer:
        return render_to_html(obj, _renderer)


def _table_cell(contents: RenderableType, props: OptionalCssProps = None, extra_class: str = '') -> str:
    cell_html = render_at_css_width(contents, props)
    return div_class(cell_html, join_truthy(extra_class, 'column'), props, role='cell')



"""
Epstein money related files. This category makes is_interesting = True.
"""
from epstein_files.documents.config.config_builder import letter
from epstein_files.documents.config.doc_cfg import DEFAULT_TRUNCATE_TO, DocCfg
from epstein_files.documents.config.email_cfg import EmailCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *


MONEY_CFGS = [
    # private placement memoranda
    DocCfg(
        id='024432',
        date='2006-09-27',
        note=f"Michael Milken's Knowledge Universe Education (KUE) $1,000,000 corporate share placement notice (SEC filing?)"
    ),
    DocCfg(id='024003', note=f"New Leaf Ventures ($375 million biotech fund) private placement memorandum"),

    # DOJ files
    DocCfg(id='EFTA00803405', author=HONEYCOMB_ASSET_MANAGEMENT, note="fund brochure", truncate_to=DEFAULT_TRUNCATE_TO),
    DocCfg(id='EFTA00006069', author='NES LLC', note='W-2 tax form issued for employee whom Epstein paid $185,323 in 2005', date='2006-01-01'),
    DocCfg(
        id='EFTA01413294',
        comment='related to EFTA01357341, efta01363125, + more based on Vavilov Street address',
        date='2016-12-16',
        date_uncertain='hard to tell',
        note='compliance report on payment from Epstein to Sberbank account of Nikolay Aleksandrovich Gyrov',
    ),
    DocCfg(
        id='EFTA00173953',
        author=OCDETF,
        note='report on investigations of Epstein related drug money laundering',
        is_interesting=10,
    ),
    DocCfg(
        id='EFTA01681865',
        author=DEUTSCHE_BANK,
        note="explanations of all of Epstein's large payments prepared for DOJ",
        is_interesting=10,
    ),
    DocCfg(
        id='EFTA01361270',
        author=DEUTSCHE_BANK,
        date='2014-01-02',
        note=f"$60,000 transfer from {SOUTHERN_TRUST_COMPANY} to {BEN_GOERTZEL}'s Novamente",
    ),
    DocCfg(
        id='EFTA01111057',
        author=MORTIMER_ZUCKERMAN,
        date='2014-07-10',
        note='Mortimer B. Zuckerman Management Trust',
        non_participants=['Marla Maples'],
    ),
    DocCfg(id='EFTA01478313', note=f'list of investments (maybe of {LEON_BLACK})', date='2016-03-31'),
    DocCfg(id='EFTA01285411', note=f"bank statement for Epstein's {SOUTHERN_TRUST_COMPANY} showing $82 million balance"),
    DocCfg(id='EFTA01222951', note=f"credit card expenses for Carlos L Rodriguez using Plum Card", date='2019-02-12'),
    DocCfg(id='EFTA00016884', note="Epstein's last will and testament", date='2014-11-18', is_interesting=10),
    DocCfg(
        id='EFTA00089546',
        note=f"Epstein last will and testament codicil naming {JAMES_CAYNE} an executor",
        date='2007-09-20',
        non_participants=[JOI_ITO],
    ),
    DocCfg(id='EFTA01266380', note="Epstein's 2014 Trust with bequests"),
    DocCfg(id='EFTA01282282', note=f"Epstein Butterfly Trust (sole beneficiary {KARYNA_SHULIAK})"),
    DocCfg(id='EFTA01583819', note=f"Epstein had control of {JAMES_CAYNE}'s assets"),
    DocCfg(id='EFTA00099424', note=f"Epstein 2017 Trust (Eva Andersson Dubin, {DARREN_INDYKE}, {RICHARD_KAHN})"),
    DocCfg(id='EFTA01266457', note=f"Epstein 2018 Trust ({KATHRYN_RUEMMLER}, {DARREN_INDYKE}, {RICHARD_KAHN})"),
    DocCfg(id='EFTA01266204', note=f"Epstein The 1953 Trust ({DARREN_INDYKE}, {RICHARD_KAHN})", date='2019-08-08'),
    DocCfg(id='EFTA00299927', note=f"estate plan for {JAMES_CAYNE} found in Epstein's possession"),
    DocCfg(id='EFTA01265973', note="large transfers around time of Epstein arrest", show_full_panel=True),
    DocCfg(id='EFTA01087311', note=f'{LEON_BLACK} Family Partners cash projections'),
    DocCfg(id='EFTA01366011', note=f"memo requesting $3,000 payment to {LASMA_KUHTARSKA}", show_with_name=LASMA_KUHTARSKA),
    DocCfg(id='EFTA01086463', note=f"{MORTIMER_ZUCKERMAN}'s art collection valuations", is_valid_for_name_scan=False),
    DocCfg(id='EFTA00007781', note='paychecks signed by Epstein deposited at Colonial Bank', date='2005-08-12'),
    DocCfg(id='EFTA01273102', note=f"payment from Epstein to {RENATA_BOLOTOVA}'s father's account at Sberbank"),
    EmailCfg(id='EFTA00994380', highlight_quote='please confirm $500 to Sergey Pozhidaev', truncate_to=500),
    DocCfg(id='EFTA00000476', display_text='photo of JEFFREY EPSTEIN CASH DISBURSEMENTS', date='2006-09-01', is_interesting=False),
    DocCfg(id='EFTA00238499', note='wire transfer to Signature Bank account'),
    DocCfg(id='EFTA00606411', display_text='proposed jet ownership structure flowchart', date='2017-01-01', date_uncertain='guess'),

    # Jeepers, Inc.
    DocCfg(id='EFTA01286368', author=DEUTSCHE_BANK, note=f'bank statements showing receipt of $2 million from {JEEPERS_INC}'),
    EmailCfg(id='EFTA01424585', note=f'{DEUTSCHE_BANK} anti-money laundering review of "high risk" {JEEPERS_INC}'),
    EmailCfg(
        id='EFTA01416658',
        highlight_quote="Southern Financial — one of the most complicated client situations I've seen",
        note=f"asking for a promotion based on work on Epstein's account",
        truncate_to=(2_600, 3_600),
    ),
    letter(
        'EFTA00591276',
        'Susman Godfrey',
        ['Fortress Investment Group'],
        highlight_quote='I represent Jeepers, Inc., Financial Trust Company, Inc., and Jeffrey Epstein',
        note="concerning Jeepers, the children's theme park owned by Epstein",
        truncate_to=(500, 1_800),
    ),

    # Emails
    EmailCfg(id='EFTA00037187', is_interesting=True),
    EmailCfg(
        id='EFTA01409449',
        note=f"{DEUTSCHE_BANK} employees scrubbing Epstein's name off his TWTR (Twitter) trades",
        is_interesting=10,
    ),
    EmailCfg(
        id='EFTA00461557',
        author=LESLEY_GROFF,
        author_uncertain='schedule',
        note=f"{ARIANE_DE_ROTHSCHILD} and {SERGEY_BELYAKOV} visiting at the same time",
    ),
    EmailCfg(
        id='EFTA02189550',
        author='Ike Groff',
        author_uncertain=True,
        note=f"Ike Groff invests $250,000 in Mangrove Partners managed by Nathaniel August",
    ),
    EmailCfg(id='EFTA01375243', note=f"{DEUTSCHE_BANK} anti-money laundering team flags Epstein's $237k transfer to {SVETLANA_POZHIDAEVA}'s family"),
    EmailCfg(id='EFTA00382048', note=f'setting up an ISDA (special account for high volume traders) with {DEUTSCHE_BANK}'),
    EmailCfg(id='EFTA00629657', note=f"arranging {LEON_BLACK}'s finances"),
    EmailCfg(id='EFTA00371120', note=f"Epstein appears to invest in {ATORUS}"),
    EmailCfg(id='EFTA00652799', note=f'Epstein calls Ari Glass "a bit sketchy" despite investing ~$50 million in his fund Boothbay'),
    EmailCfg(id='EFTA01388422', note='Nadean Novogratz is probably the sister-in-law of crypto ponzi billionaire Mike Novogratz'),
]
